# Imports from the python standard library to build the plugin functionality
import os.path


from UM.i18n import i18nCatalog
from cura.CuraApplication import CuraApplication

# Importe von QT für Signale and Slots
from PyQt5.QtCore import QObject, pyqtProperty, pyqtSignal, pyqtSlot
from PyQt5.QtWidgets import QLabel, QApplication # brauchen wir das?

# Importe from Uranium and Cura to interact with the plugin system
from UM.Extension import Extension
from UM.Application import Application
from UM.PluginRegistry import PluginRegistry
from UM.Scene.Selection import Selection


# Importe von Uranium, um die Internationalization freizugeben
from UM.i18n import i18nCatalog
i18n_catalog = i18nCatalog("Serializer")


from . import PrinterInfo


# Klasse für das Plugin Serializer, einer Extension in Cura
class Serializer(QObject, Extension):

    # Definiere ein pyqtSignal
    valuesChanged = pyqtSignal()

# Konstuktor ----------------------------------------------------------------------------------------------------------
    # registriert das Menu Item, initialisiert die Instanzvariablen und
    # verbindet den Seriendruck mit dem Signal writeStarted des OutputDeviceManagers

    def __init__(self, parent=None) -> None:
        QObject.__init__(self, parent)
        Extension.__init__(self)

        # Menüeinträge
        self.setMenuName(i18n_catalog.i18nc("@item:inmenu", "Serializer"))
        self.addMenuItem(i18n_catalog.i18nc("@item:inmenu", "Open Serializer"), self.showSerializer)
        self._serializer_window = None
        self._infill_methode = 3 #Methoden haben Index 0 bis 4
        self._fuelldichte = 50.0
        self._stueckzahl = 1
        self._qualitaet = 10
        # speichert, ob der fertig-Button gedrückt wurde, 1 - gedrückt, 0 - nicht gedrückt
        self._fertigclicked = 0
        # speichert den Namen des Objekts auf der aktuellen Buildplate
        self._objektName = ""
        # speichert Drucker-spezifische Informationen
        self._printerInfo = PrinterInfo.PrinterInfo()


        # Die Funktion zum Seriendruck wird ausgeführt, wenn emittiert wird, dass in ein Outputdevice
        # geschrieben werden soll
        Application.getInstance().getOutputDeviceManager().writeStarted.connect(self.seriesprint)


# System Anbindung ----------------------------------------------------------------------------------------------------
    # Wenn das Sarlex-Serializer Fenster bereits geladen wurde, wird das Fenster beim Klicken
    # auf das Menuitem geöffnet
    def showSerializer(self):
        if self._serializer_window is None:
            self._serializer_window = self._createDialogue()
        self._serializer_window.show()

    # Methode, die das Fenster anhand der qml-Datei erstellt, das Serializer-Objekt als manager Variable registriert
    # und neben dem Slice-Button eine Notification erzeugt
    def _createDialogue(self):
        # Pfad zur qml-Datei angeben
        qml_file_path = os.path.join(PluginRegistry.getInstance().getPluginPath(self.getPluginId()), "SerializerUI.qml")
        # Gui Instanz erstellen und den manager festlegen zur Interaktion
        component_with_context = Application.getInstance().createQmlComponent(qml_file_path, {"manager": self})
        # Notification erstellen
        CuraApplication.getInstance().addAdditionalComponent("saveButton", component_with_context.findChild(QObject, "SerializerStuekzahlButton"))
        return component_with_context


# Input und Verrechnung von Werten ------------------------------------------------------------------------------------

    # speichert, dass der fertig-Button gedrückt wurde
    @pyqtSlot()
    def update_fertig_button(self):
        self._fertigclicked = 1
        self.valuesChanged.emit()

    # wenn der Drucker gewechselt wurde, ändern sich die Infos zu ihm
    @pyqtSlot(str)
    def updatemachine(self, machinename):
        self._printerInfo.setPrinterName(machinename)

    # wenn sich x- oder z-Koordinate eines Objekts ändern, wird dies dem Printerinfo-Objekt mitgeteilt
    @pyqtSlot(str)
    def updatex(self, var_x):
        x = 0
        try:
            x = round(float(var_x), 4)
        except Exception as e:
            x = 0
        self._printerInfo.setxcoordinate(x)

    # Stückzahl ändert sich,
    # Exception: sobald das Feld leer ist, gibt es einen Fehler, weil das leere Wort nicht in float konvertiert werden kann
    # + nur ganze Zahlen werden weiterverarbeitet
    @pyqtSlot(str)
    def update_stueckzahl(self, var_stueckzahl):
        try:
            self._stueckzahl = int(var_stueckzahl)
        except Exception as e:
            pass
        self.valuesChanged.emit()


    # Qualitätänderung:
    # Qualität und Fülldichte verhalten sich proportional
    # Bei Änderung der Qualität ändert sich die Fülldichte
    # Approximation: Dichte = Qualität / ((1/6) * k-Value von einer Methode)
    @pyqtSlot(int)
    def update_qualitaet(self, var_qualitaet):
        self._qualitaet = var_qualitaet
        self.calculate_with_qualitaet()

    def calculate_with_qualitaet(self):
        self._fuelldichte = self._qualitaet / ((1 / 6) * self.kvalue(self._infill_methode))
        self.valuesChanged.emit()

    # Hilfsmethoden
    # Konversion zwischen Infillmethodenindex und k-Value
    # der k-Value orientiert sich an der Anzahl der Kanten eines Körpers
    # k-Value ~ Qualität
    # Ordnung nach k-Value: Viereck > ZigZag > Hexagon > Triangle > Grid
    def kvalue(self, index):
        index = int(index)
        if index == 0:
            return 12
        elif index == 1:
            return 6
        elif index == 2:
            return 8
        elif index == 3:
            return 7
        elif index == 4:
            return 10
        else:
            return 0

    # Infillmethodenänderung: Änderung der Qualität
    # Qualität = Dichte * ((1/6) * k-Value von einer Methode

    @pyqtSlot(str)
    def update_infill_methode(self, var_methode):
        self._infill_methode = var_methode
        self.calculate_with_infill_methode()

    def calculate_with_infill_methode(self):
        self._qualitaet = self._fuelldichte * (1 / 6) * self.kvalue(self._infill_methode)
        self.valuesChanged.emit()

    # Infill-Stärke ändert sich: Qualität ändert sich
    # Qualität = Dichte * ((1/6) * k-Value von einer Methode

    @pyqtSlot(str)
    def update_fuelldichte(self, var_fuelldichte):
        self._fuelldichte = float(var_fuelldichte)
        self.calculate_with_fuelldichte()

    def calculate_with_fuelldichte(self):
        self._qualitaet = self._fuelldichte * (1 / 6) * self.kvalue(self._infill_methode)
        self.valuesChanged.emit()


# Funktionen, die der GUI den aktuellen Wert geben --------------------------------------------------------------------
    @pyqtProperty(int, notify=valuesChanged)
    def fuelldichte(self):
        return self._fuelldichte

    @pyqtProperty(int, notify=valuesChanged)
    def stueckzahl(self):
        return self._stueckzahl

    @pyqtProperty(int, notify=valuesChanged)
    def qualitaet(self):
        return self._qualitaet

    @pyqtProperty(str, notify=valuesChanged)
    def objektName(self):
        try:
            self._objektName = Selection.getSelectedObject(0).getName()
        except Exception as e:
            self._objektName = "Wählen Sie bitte ein Objekt aus."
        return self._objektName

    @pyqtProperty(int, notify=valuesChanged)
    def fertigclicked(self):
        return self._fertigclicked

# Seriendruck ---------------------------------------------------------------------------------------------------------
    # gcode vervielfachen und Objekte herunterschubsen
    # Quelle für die gcode Befehle: https://github.com/Ultimaker/Marlin/blob/Marlin_v1/README.md und
    # https://marlinfw.org/meta/gcode/

    def seriesprint(self, output_device):

        # nur wenn ein Seriendruck angefordert wurde, wird der gcode verändert
        if self._fertigclicked == 0:
            return

        # Hole den gcode zum Objekt, wenn es keinen gcode gibt, tu nichts
        scene = Application.getInstance().getController().getScene()
        if not hasattr(scene, "gcode_dict"):
            return
        gcode_dict = getattr(scene, "gcode_dict")
        if not gcode_dict:
            return
        active_build_plate_id = CuraApplication.getInstance().getMultiBuildPlateModel().activeBuildPlate
        gcode_list = gcode_dict[active_build_plate_id]
        if not gcode_list:
            return

        # Wenn die Stückzahl 0 ist, braucht nichts gedruckt zu werden
        if self._stueckzahl == 0:
            gcode_list = [";Stückzahl = 0! Es ist nichts zu drucken. \n"]
            gcode_dict[active_build_plate_id] = gcode_list
            setattr(scene, "gcode_dict", gcode_dict)
            return

        # Schreibe als Kommentar, dass es sich um einen Seriendruck handelt und wieviele Objekte gedruckt werden
        if self._stueckzahl > 1:
            gcode_list[0] += ";Series print: " + str(self._stueckzahl) + " objects\n"
        else:
            gcode_list[0] += ";Series print: " + str(self._stueckzahl) + " object\n"

        # ersten Paragraphen und letzten Paragraph herausschneiden und zwischenspeichern
        firstparagraph = gcode_list.pop(0)
        lastparagraph = gcode_list.pop()

        # ersten Paragraphen an das PrinterInfo-Objekt übergeben
        self._printerInfo.setFirstParagraph(firstparagraph)

        # Die Koordinaten für den Extruder zum Herunterschubsen finden
        backy = self._printerInfo.findbacky()
        pushx = self._printerInfo.findpushx()
        pushy = self._printerInfo.findpushy()
        pushz = self._printerInfo.findpushz()

        #Herunterschubsen
        # Strategie: Der Extruder befindet sich nach dem Druck über dem Objekt. Er geht dann in y-Richtung nach hinten.
        # Dann bewegt er sich in die Mitte des Objekts schrittweise: erst in x-Richtung, dann in z-Richtung.
        # Dann wartet er ggf und das Heizbett wird abgekühlt. Der Extruder fährt bis zum maximalen Anschlag in y-Richtung
        # Ist noch ein Druck nötig, wird wieder aufgeheizt.

        gcode_list[-1] += ";HERUNTERSCHUBSEN\n"
        gcode_list[-1] += "G0 Y" + backy + "\n"
        gcode_list[-1] += "G0 X" + str(pushx) + "\n"
        gcode_list[-1] += "G0 Z" + str(pushz) + "\n"
        gcode_list[-1] += self._printerInfo.waitncool()
        # Rausschieben mit minimaler Geschwindikeit des Druckers
        gcode_list[-1] += "G0 Y" + pushy + self._printerInfo.slowestspeed()
        gcode_list[-1] += ";HERUNTERSCHUBSEN\n"

        # Code vervielfachen
        copygcode = gcode_list.copy()
        gcode_list[0] += "; 1.object\n"
        for i in range(1, self._stueckzahl):
            gcode_list = gcode_list + [";" + str(i+1) + ". object\n"]
            gcode_list = gcode_list + copygcode

        # ersten und letzten Paragraphen des gcodes wieder hinzufuegen
        gcode_list = [firstparagraph] + gcode_list
        gcode_list = gcode_list + [lastparagraph]

        # den manipulierten gcode an das System übergeben
        gcode_dict[active_build_plate_id] = gcode_list
        setattr(scene, "gcode_dict", gcode_dict)

# Seriendruck Ende-----------------------------------------------------------------------------------------------------



